# -*- coding: utf-8 -*-
import logging

import septentrion

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections
from django.db import DEFAULT_DB_ALIAS

from django_north.management.commands import septentrion_settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Migrate the DB to the target version."

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput', '--no-input', action='store_false',
            dest='interactive',
            help='Tells Django to NOT prompt the user for input of any kind.',
        )
        parser.add_argument(
            '--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. '
                 'Defaults to the "default" database.',
        )
        parser.add_argument(
            '--run-syncdb', action='store_true', dest='run_syncdb',
            help='Creates tables for apps without migrations.',
        )

    def handle(self, *args, **options):
        if getattr(settings, 'NORTH_MANAGE_DB', False) is not True:
            logger.info('migrate command disabled')
            return

        self.verbosity = options.get('verbosity')

        connection = connections[options['database']]
        septentrion.migrate(**septentrion_settings(connection))



from django.conf import settings


def septentrion_settings(connection):

    settings_dict = {}
    mapping = {
        "migrations_root": "NORTH_MIGRATIONS_ROOT",
        "target_version": "NORTH_TARGET_VERSION",
        "schema_template": "NORTH_SCHEMA_TPL",
        "schema_version": "NORTH_SCHEMA_VERSION",
        "target_version": "NORTH_TARGET_VERSION",
        "fixtures_template": "NORTH_FIXTURES_TPL",
        "before_schema_file": "NORTH_BEFORE_SCHEMA_FILES",
        "after_schema_file": "NORTH_AFTER_SCHEMA_FILES",
        "non_transactional_keyword": "NORTH_NON_TRANSACTIONAL_KEYWORDS",
        "target_version": "NORTH_TARGET_VERSION",
    }

    for septentrion_name, django_name in mapping.items():
        try:
            settings_dict[septentrion_name] = getattr(settings, django_name)
        except AttributeError:
            pass

    settings_dict.update({key: value for key, value in {
        # Settings from Django's DB connection
        "dbname": connection.settings_dict["NAME"],
        "host": connection.settings_dict["HOST"],
        "port": connection.settings_dict["PORT"],
        "username": connection.settings_dict["USER"],
        "password": connection.settings_dict["PASSWORD"],
    }.items() if value})

    settings_dict.update({
        # Static settings
        "table": "django_migrations",
        "version_column": "app",
        "name_column": "name",
        "applied_at_column": "applied",
        "create_table": False,
    })
    return settings_dict



import os
from distutils.version import StrictVersion
from importlib import import_module

from django.conf import settings
from django.db.migrations.recorder import MigrationRecorder
from django.db.utils import ProgrammingError

import septentrion


fixtures_default_tpl = 'fixtures_{}.sql'
schema_default_tpl = 'schema_{}.sql'


class DBException(Exception):
    pass


def is_version(vstring):
    try:
        StrictVersion(vstring)
    except ValueError:
        return False
    return True


def list_dirs(root):
    return [d for d in os.listdir(root)
            if os.path.isdir(os.path.join(root, d))]


def list_files(root):
    return [d for d in os.listdir(root)
            if os.path.isfile(os.path.join(root, d))]


def get_applied_versions(connection):
    """
    Return the list of applied versions.
    Reuse django migration table.
    """
    recorder = MigrationRecorder(connection)
    applied_versions = list(recorder.migration_qs.filter(
        app__in=septentrion.get_known_versions(
            MIGRATIONS_ROOT=settings.NORTH_MIGRATIONS_ROOT,
        )
    ).values_list(
        'app', flat=True).order_by('app').distinct())

    # sort versions
    applied_versions.sort(key=StrictVersion)
    return applied_versions


def get_current_version(connection):
    """
    Return the current version of the database.
    Return None if the schema is not inited.
    """
    try:
        import_string = settings.NORTH_CURRENT_VERSION_DETECTOR
    except AttributeError:
        import_string = (
            'django_north.management.migrations'
            '.get_current_version_from_table')

    module_path, factory_name = import_string.rsplit('.', 1)
    module = import_module(module_path)
    factory = getattr(module, factory_name)

    return factory(connection)


def get_current_version_from_table(connection):
    """
    Return the current version of the database, from sql_version table.
    Return None if the table does not exist (schema not inited).
    """
    with connection.cursor() as cursor:
        try:
            cursor.execute("SELECT version_num FROM sql_version;")
        except ProgrammingError:
            # table does not exist ?
            return None

        rows = cursor.fetchall()

    versions = [row[0] for row in rows if is_version(row[0])]
    if not versions:
        return None

    # sort versions
    versions.sort(key=StrictVersion)

    # return the last one
    return versions[-1]


def get_current_version_from_comment(connection):
    """
    Return the current version of the database, from django_site comment.
    Return None if the table django_site does not exist (schema not inited).
    """
    with connection.cursor() as cursor:
        try:
            cursor.execute(
                "SELECT obj_description('django_site'::regclass, 'pg_class');")
        except ProgrammingError:
            # table does not exist ?
            return None

        row = cursor.fetchone()
        comment = row[0]

    # no comment
    if comment is None:
        raise DBException('No comment found on django_site.')

    # parse comment
    if 'version ' not in comment:
        raise DBException("No version found in django_site's comment.")
    return comment.replace('version ', '').strip()


def get_applied_migrations(version, connection):
    """
    Return the list of applied migrations for the given version.
    Reuse django migration table.
    """
    recorder = MigrationRecorder(connection)
    return list(recorder.migration_qs.filter(app=version).values_list(
        'name', flat=True))



from __future__ import unicode_literals

from django.core.management.base import AppCommand
from django.core.management.base import CommandError
from django.db import DEFAULT_DB_ALIAS, connections
from django.db import router
from django.utils.version import get_docs_version


def sql_create_model(editor, model):
    """
    Takes a model and creates a table for it in the database.
    Will also create any accompanying indexes or unique constraints.
    """
    # Create column SQL, add FK deferreds if needed
    column_sqls = []
    params = []
    for field in model._meta.local_fields:
        # SQL
        definition, extra_params = editor.column_sql(model, field)
        if definition is None:
            continue
        # Check constraints can go on the column SQL here
        db_params = field.db_parameters(connection=editor.connection)
        if db_params['check']:
            definition += " CHECK (%s)" % db_params['check']
        # Autoincrement SQL (for backends with inline variant)
        col_type_suffix = field.db_type_suffix(connection=editor.connection)
        if col_type_suffix:
            definition += " %s" % col_type_suffix
        params.extend(extra_params)
        # FK
        remote_field = field.remote_field
        if remote_field and field.db_constraint:
            editor.deferred_sql.append(
                editor._create_fk_sql(
                    model, field, "_fk_%(to_table)s_%(to_column)s"))
        # Add the SQL to our big list
        column_sqls.append("%s %s" % (
            editor.quote_name(field.column),
            definition,
        ))
        # Autoincrement SQL (for backends with post table definition variant)
        if field.get_internal_type() == "AutoField":
            autoinc_sql = editor.connection.ops.autoinc_sql(
                model._meta.db_table, field.column)
            if autoinc_sql:
                editor.deferred_sql.extend(autoinc_sql)

    # Add any unique_togethers (always deferred, as some fields might be
    # created afterwards, like geometry fields with some backends)
    for fields in model._meta.unique_together:
        columns = [model._meta.get_field(field).column for field in fields]
        editor.deferred_sql.append(editor._create_unique_sql(model, columns))
    # Make the table
    sql = editor.sql_create_table % {
        "table": editor.quote_name(model._meta.db_table),
        "definition": ", ".join(column_sqls)
    }
    if model._meta.db_tablespace:
        tablespace_sql = editor.connection.ops.tablespace_sql(
            model._meta.db_tablespace)
        if tablespace_sql:
            sql += ' ' + tablespace_sql
    # Prevent using [] as params, in the case a literal '%' is used in the
    # definition
    output = [sql % params]

    # Add any field index and index_together's
    editor.deferred_sql.extend(editor._model_indexes_sql(model))

    return output


def sql_create(app_config, style, connection):
    "Returns a list of the CREATE TABLE SQL statements for the given app."

    if connection.settings_dict['ENGINE'] == 'django.db.backends.dummy':
        # This must be the "dummy" database backend, which means the user
        # hasn't set ENGINE for the database.
        raise CommandError(
            "Django doesn't know which syntax to use for your SQL statements,"
            "\nbecause you haven't properly specified the ENGINE setting for "
            "the database.\n"
            "see: https://docs.djangoproject.com/en/%s/ref/settings/"
            "#databases" % get_docs_version()
        )

    # Get installed models, so we generate REFERENCES right.
    # We trim models from the current app so that the sqlreset command does not
    # generate invalid SQL (leaving models out of known_models is harmless, so
    # we can be conservative).
    final_output = []
    output = []

    for model in router.get_migratable_models(
            app_config, connection.alias, include_auto_created=True):
        with connection.schema_editor() as editor:
            output += sql_create_model(editor, model)
            final_output.extend(editor.deferred_sql)
            editor.deferred_sql = []
    if not get_docs_version().startswith('1'):
        # Content of final_output are Statement objects which need to be
        # converted as str
        final_output = [str(statement) for statement in final_output]
    return output + final_output


def sql_all(app_config, style, connection):
    """
    Returns a list of CREATE TABLE SQL, initial-data inserts,
    and CREATE INDEX SQL for the given module.
    """
    return sql_create(app_config, style, connection)


class Command(AppCommand):
    help = ("Prints the CREATE TABLE, custom SQL and CREATE INDEX SQL "
            "statements for the given model module name(s).")

    output_transaction = True

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            '--database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to print the SQL for. Defaults to the '
                 '"default" database.')

    def handle_app_config(self, app_config, **options):
        if app_config.models_module is None:
            return
        connection = connections[options['database']]
        statements = sql_all(app_config, self.style, connection)
        return '\n'.join(statements)



# -*- coding: utf-8 -*-
import logging

import septentrion

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections
from django.db import DEFAULT_DB_ALIAS

from django_north.management.commands import septentrion_settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Shows all available migrations for the current project"

    def add_arguments(self, parser):
        parser.add_argument(
            '--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to synchronize. '
                 'Defaults to the "default" database.',
        )

    def handle(self, *args, **options):
        if getattr(settings, 'NORTH_MANAGE_DB', False) is not True:
            logger.info('showmigrations command disabled')
            return

        connection = connections[options['database']]

        septentrion.show_migrations(**septentrion_settings(connection),)
